@@ -1,14 +1,3 @@
-/******************************************************************************
-*******************************************************************************
-**
-**  Copyright (C) 2005 Red Hat, Inc.  All rights reserved.
-**
-**  This copyrighted material is made available to anyone wishing to use,
-**  modify, copy, or redistribute it subject to the terms and conditions
-**  of the GNU General Public License v.2.
-**
-*******************************************************************************
-******************************************************************************/
 #include <stdio.h>
 #include <errno.h>
 #include <string.h>
@@ -67,7 +56,7 @@ static int get_doc_version(xmlDocPtr ldoc)
     goto fail;
   }
 
-  obj = xmlXPathEvalExpression((xmlChar *)"//cluster/@config_version", ctx);
+  obj = xmlXPathEvalExpression((xmlChar *)"/cluster/@config_version", ctx);
 
   if (!obj || !obj->nodesetval || (obj->nodesetval->nodeNr != 1)) {
     fprintf(stderr, "Error while retrieving config_version.\n");
@@ -150,7 +139,7 @@ int update(char *location)
     return desc;
   }
 
-  if ((error = ccs_get(desc, "//@config_version", &v1_str))) {
+  if ((error = ccs_get(desc, "/cluster/@config_version", &v1_str))) {
     fprintf(stderr, "Unable to get current config_version: %s\n", strerror(-error));
     ccs_disconnect(desc);
     return error;
@@ -355,7 +344,7 @@ int update(char *location)
     return 0;
   }
 
-  if ((error = ccs_get(desc, "//@config_version", &v3_str))) {
+  if ((error = ccs_get(desc, "/cluster/@config_version", &v3_str))) {
     ccs_disconnect(desc);
     return 0;
   }
@@ -379,7 +368,7 @@ int update(char *location)
 	    "  Old config version :: %d\n"
 	    "  Proposed config version :: %d\n"
 	    "  Winning config version  :: %d\n\n"
-	    "Check /etc/cluster/cluster.conf to ensure it contains the desired contents.\n", v1, v2, v3);
+	    "Check " DEFAULT_CONFIG_DIR "/" DEFAULT_CONFIG_FILE " to ensure it contains the desired contents.\n", v1, v2, v3);
     return -EAGAIN;
   }
 
