@@ -219,6 +219,7 @@ static void show_status(void)
 	if (cman_get_quorum_device(h, &qinfo) == 0 && qinfo.qi_state == 2)
 		printf("Quorum device votes: %d\n", qinfo.qi_votes);
 	printf("Total votes: %d\n", einfo->ei_total_votes);
+	printf("Node votes: %d\n", einfo->ei_node_votes);
 
 	printf("Quorum: %d %s\n", einfo->ei_quorum, quorate?" ":"Activity blocked");
 	printf("Active subsystems: %d\n", cman_get_subsys_count(h));
@@ -553,11 +554,18 @@ static void leave(commandline_t *comline)
 	}
 
 	result = cman_shutdown(h, flags);
-	if (result) {
+	if (result && !comline->wait_opt) {
 		die("Error leaving cluster: %s", cman_error(errno));
 	}
-
 	cman_finish(h);
+
+	/* Wait until cman shuts down */
+	if (comline->wait_opt) {
+		while ( (h = cman_admin_init(NULL)) ) {
+			cman_finish(h);
+			sleep(1);
+		}
+	}
 }
 
 static void set_expected(commandline_t *comline)
