@@ -609,49 +609,49 @@ static int get_ccs_join_info(void)
 		goto out;
 	}
 
-	/* Sum node votes for expected */
-	if (expected_votes == 0) {
-		for (i = 1; ; i++) {
-			int ret;
+	/* Sum node votes for expected. Even if we already know expected_votes, we need votes_sum
+	   later */
+	for (i = 1; ; i++) {
+		int ret;
 
-			name = NULL;
-			ret = snprintf(path, sizeof(path), NODE_NAME_PATH_BYNUM, i);
-			if (ret < 0 || (size_t) ret >= sizeof(path)) {
-				error = -E2BIG;
-				break;
-			}
-
-			error = ccs_get(cd, path, &name);
-			if (error || !name)
-				break;
-
-			node_count++;
-
-			ret = snprintf(path, sizeof(path), NODE_VOTES_PATH, name);
-			free(name);
-
-			if (ret < 0 || (size_t) ret >= sizeof(path)) {
-				error = -E2BIG;
-				break;
-			}
-
-			error = ccs_get(cd, path, &str);
-			if (error)
-				vote_sum++;
-			else {
-				if (atoi(str) < 0) {
-					log_printf(LOG_ERR, "negative votes not allowed");
-					write_cman_pipe("Found negative votes for this node in CCS");
-					error = -EINVAL;
-					goto out;
-				}
-				vote_sum += atoi(str);
-				free(str);
-			}
+		name = NULL;
+		ret = snprintf(path, sizeof(path), NODE_NAME_PATH_BYNUM, i);
+		if (ret < 0 || (size_t) ret >= sizeof(path)) {
+			error = -E2BIG;
+			break;
 		}
 
-		/* optional expected_votes supercedes vote sum */
+		error = ccs_get(cd, path, &name);
+		if (error || !name)
+			break;
 
+		node_count++;
+
+		ret = snprintf(path, sizeof(path), NODE_VOTES_PATH, name);
+		free(name);
+
+		if (ret < 0 || (size_t) ret >= sizeof(path)) {
+			error = -E2BIG;
+			break;
+		}
+
+		error = ccs_get(cd, path, &str);
+		if (error)
+			vote_sum++;
+		else {
+			if (atoi(str) < 0) {
+				log_printf(LOG_ERR, "negative votes not allowed");
+				write_cman_pipe("Found negative votes for this node in CCS");
+				error = -EINVAL;
+				goto out;
+			}
+			vote_sum += atoi(str);
+			free(str);
+		}
+	}
+
+	/* optional expected_votes supercedes vote sum */
+	if (expected_votes == 0) {
 		error = ccs_get(cd, EXP_VOTES_PATH, &str);
 		if (!error) {
 			expected_votes = atoi(str);
