@@ -1,21 +1,8 @@
-/******************************************************************************
-*******************************************************************************
-**
-**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
-**
-**  This copyrighted material is made available to anyone wishing to use,
-**  modify, copy, or redistribute it subject to the terms and conditions
-**  of the GNU General Public License v.2.
-**
-*******************************************************************************
-******************************************************************************/
-
 #include <linux/sched.h>
 #include <linux/slab.h>
 #include <linux/smp_lock.h>
 #include <linux/spinlock.h>
-#include <asm/semaphore.h>
+#include <linux/semaphore.h>
 #include <linux/completion.h>
 #include <linux/buffer_head.h>
 #include <linux/posix_acl.h>
@@ -170,6 +157,7 @@ gfs_inode_attr_out(struct gfs_inode *ip)
 struct inode *
 gfs_iget(struct gfs_inode *ip, int create)
 {
+	struct gfs_sbd *sdp = ip->i_sbd;
 	struct inode *inode = NULL, *tmp;
 
 	spin_lock(&ip->i_spin);
@@ -189,11 +177,26 @@ gfs_iget(struct gfs_inode *ip, int create)
 	/* Attach GFS-specific ops vectors */
 	if (ip->i_di.di_type == GFS_FILE_REG) {
 		tmp->i_op = &gfs_file_iops;
-		tmp->i_fop = &gfs_file_fops;
-		tmp->i_mapping->a_ops = &gfs_file_aops;
+		memcpy(&ip->gfs_file_aops, &gfs_file_aops,
+			   sizeof(struct address_space_operations));
+		tmp->i_mapping->a_ops = &ip->gfs_file_aops;
+		if (sdp->sd_args.ar_localflocks) {
+			if (gfs_is_jdata(ip))
+				tmp->i_fop = &gfs_file_fops_nolock_jdata;
+			else
+				tmp->i_fop = &gfs_file_fops_nolock;
+		} else {
+			if (gfs_is_jdata(ip))
+				tmp->i_fop = &gfs_file_fops_jdata;
+			else
+				tmp->i_fop = &gfs_file_fops;
+		}
 	} else if (ip->i_di.di_type == GFS_FILE_DIR) {
 		tmp->i_op = &gfs_dir_iops;
-		tmp->i_fop = &gfs_dir_fops;
+		if (sdp->sd_args.ar_localflocks)
+			tmp->i_fop = &gfs_dir_fops_nolock;
+		else
+			tmp->i_fop = &gfs_dir_fops;
 	} else if (ip->i_di.di_type == GFS_FILE_LNK) {
 		tmp->i_op = &gfs_symlink_iops;
 	} else {
@@ -338,6 +341,10 @@ inode_create(struct gfs_glock *i_gl, struct gfs_inum *inum,
 	gfs_glock_hold(i_gl);
 	set_gl2ip(i_gl, ip);
 
+	/* initialize stat counter and timestamp */
+	ip->i_dir_stats = 0;
+	do_gettimeofday(&ip->i_dir_stat_st);
+
 	atomic_inc(&sdp->sd_inode_count);
 
 	*ipp = ip;
@@ -799,7 +806,7 @@ inode_dealloc_uninit(struct gfs_sbd *sdp, struct gfs_inum *inum)
  fail:
 	gfs_glock_dq_uninit(&ri_gh);
 
-	return error;	
+	return error;
 }
 
 /**
@@ -914,7 +921,7 @@ gfs_lookupi(struct gfs_holder *d_gh, struct qstr *name,
 		return error;
 
 	if (!is_root) {
-		error = permission(dip->i_vnode, MAY_EXEC, NULL);
+		error = inode_permission(dip->i_vnode, MAY_EXEC);
 		if (error) {
 			gfs_glock_dq(d_gh);
 			return error;
@@ -956,7 +963,7 @@ gfs_lookupi(struct gfs_holder *d_gh, struct qstr *name,
 		}
 
 		if (!is_root) {
-			error = permission(dip->i_vnode, MAY_EXEC, NULL);
+			error = inode_permission(dip->i_vnode, MAY_EXEC);
 			if (error) {
 				gfs_glock_dq(d_gh);
 				gfs_glock_dq_uninit(i_gh);
@@ -1021,7 +1028,7 @@ create_ok(struct gfs_inode *dip, struct qstr *name, unsigned int type)
 {
 	int error;
 
-	error = permission(dip->i_vnode, MAY_WRITE | MAY_EXEC, NULL);
+	error = inode_permission(dip->i_vnode, MAY_WRITE | MAY_EXEC);
 	if (error)
 		return error;
 
@@ -1581,7 +1588,7 @@ gfs_unlink_ok(struct gfs_inode *dip, struct qstr *name, struct gfs_inode *ip)
 	if (IS_APPEND(dip->i_vnode))
 		return -EPERM;
 
-	error = permission(dip->i_vnode, MAY_WRITE | MAY_EXEC, NULL);
+	error = inode_permission(dip->i_vnode, MAY_WRITE | MAY_EXEC);
 	if (error)
 		return error;
 
