@@ -1100,11 +1100,12 @@ static int init(struct gfs2_sbd *sbp)
 	sbp->sd_jbsize = sbp->bsize - sizeof(struct gfs2_meta_header);
 	brelse(bh, not_updated);
 	sbp->sd_max_height = compute_heightsize(sbp, sbp->sd_heightsize,
-						sbp->sd_diptrs,
+						sbp->bsize, sbp->sd_diptrs,
 						sbp->sd_inptrs);
 	sbp->sd_max_jheight = compute_heightsize(sbp, sbp->sd_jheightsize,
-						sbp->sd_diptrs,
-						sbp->sd_inptrs);
+						 sbp->sd_jbsize,
+						 sbp->sd_diptrs,
+						 sbp->sd_inptrs);
 	/* -------------------------------------------------------- */
 	/* Our constants are for gfs1.  Need some for gfs2 as well. */
 	/* -------------------------------------------------------- */
@@ -1112,9 +1113,11 @@ static int init(struct gfs2_sbd *sbp)
                 sizeof(uint64_t); /* How many ptrs can we fit on a block? */
 	memset(gfs2_heightsize, 0, sizeof(gfs2_heightsize));
 	gfs2_max_height = compute_heightsize(sbp, gfs2_heightsize,
+					     sbp->bsize,
 					     sbp->sd_diptrs, gfs2_inptrs);
 	memset(gfs2_jheightsize, 0, sizeof(gfs2_jheightsize));
 	gfs2_max_jheight = compute_heightsize(sbp, gfs2_jheightsize,
+					      sbp->sd_jbsize,
 					      sbp->sd_diptrs, gfs2_inptrs);
 
 	/* ---------------------------------------------- */
@@ -1356,6 +1359,7 @@ int journ_space_to_rg(struct gfs2_sbd *sdp)
 		rgd->ri.ri_bitbytes = rgd->ri.ri_data / GFS2_NBBY;
 		convert_bitmaps(sdp, rgd, FALSE); /* allocates rgd->bh */
 		for (x = 0; x < rgd->ri.ri_length; x++) {
+			rgd->bh[x]->b_count++;
 			if (x)
 				gfs2_meta_header_out(&mh, rgd->bh[x]->b_data);
 			else
