@@ -295,11 +295,11 @@ static int fill_super_block(struct gfs2_sbd *sdp)
 	 ********************************************************************/
 	log_info("Initializing lists...\n");
 	osi_list_init(&sdp->rglist);
-	osi_list_init(&sdp->buf_list);
+	init_buf_list(sdp, &sdp->buf_list, 128 << 20);
+	init_buf_list(sdp, &sdp->nvbuf_list, 0xffffffff);
 	for(i = 0; i < BUF_HASH_SIZE; i++) {
 		osi_list_init(&dir_hash[i]);
 		osi_list_init(&inode_hash[i]);
-		osi_list_init(&sdp->buf_hash[i]);
 	}
 
 	/********************************************************************
