@@ -62,7 +62,7 @@ int check_eattr_indir(struct gfs2_inode *ip, uint64_t block,
 		return 1;
 	}
 	else
-		indir_bh = bread(sbp, block);
+		indir_bh = bread(&sbp->buf_list, block);
 
 	*bh = indir_bh;
 	return 0;
@@ -97,7 +97,7 @@ int check_eattr_leaf(struct gfs2_inode *ip, uint64_t block,
 		return 1;
 	}
 	else 
-		*bh = bread(sbp, block);
+		*bh = bread(&sbp->buf_list, block);
 
 	return 0;
 }
@@ -243,7 +243,7 @@ int pass1c(struct gfs2_sbd *sbp)
 
 		if (skip_this_pass || fsck_abort) /* if asked to skip the rest */
 			return 0;
-		bh = bread(sbp, block_no);
+		bh = bread(&sbp->buf_list, block_no);
 		if (gfs2_check_meta(bh, GFS2_METATYPE_IN)) { /* if a dinode */
 			log_info("EA in inode %"PRIu64" (0x%" PRIx64 ")\n",
 				 block_no, block_no);
