@@ -377,6 +377,7 @@ void *gfs2_block_list_destroy(struct gfs2_sbd *sdp,
 
 /* buf.c */
 void init_buf_list(struct gfs2_sbd *sdp, struct buf_list *bl, uint32_t limit);
+struct gfs2_buffer_head *bfind(struct buf_list *bl, uint64_t num);
 struct gfs2_buffer_head *bget_generic(struct buf_list *bl, uint64_t num,
 				      int find_existing, int read_disk);
 struct gfs2_buffer_head *bget(struct buf_list *bl, uint64_t num);
@@ -631,7 +632,7 @@ int gfs2_query(int *setonabort, struct gfs2_options *opts,
 #define SYS_BASE "/sys/fs/gfs2"
 
 uint32_t compute_heightsize(struct gfs2_sbd *sdp, uint64_t *heightsize,
-			    int diptrs, int inptrs);
+			    uint32_t bsize1, int diptrs, int inptrs);
 void compute_constants(struct gfs2_sbd *sdp);
 int find_gfs2_meta(struct gfs2_sbd *sdp);
 int dir_exists(const char *dir);
