@@ -1,16 +1,3 @@
-/******************************************************************************
-*******************************************************************************
-**
-**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
-**
-**  This copyrighted material is made available to anyone wishing to use,
-**  modify, copy, or redistribute it subject to the terms and conditions
-**  of the GNU General Public License v.2.
-**
-*******************************************************************************
-******************************************************************************/
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -29,10 +16,10 @@
 #define __user
 #include <linux/gfs2_ondisk.h>
 #include <sys/mount.h>
-#include <linux/fs.h>
 
 #include "libgfs2.h"
 #include "gfs2_tool.h"
+#include "iflags.h"
 
 #if GFS2_TOOL_FEATURE_IMPLEMENTED
 /**
@@ -115,27 +102,26 @@ print_lockdump(int argc, char **argv)
 	char *name, line[PATH_MAX];
 	char *debugfs;
 	FILE *file;
-	int rc = -1, debug_dir_existed = 1;
+	int rc = -1;
 
 	/* See if debugfs is mounted, and if not, mount it. */
 	debugfs = find_debugfs_mount();
 	if (!debugfs) {
-		debugfs = malloc(20);
+		debugfs = malloc(PATH_MAX);
 		if (!debugfs)
 			die("Can't allocate memory for debugfs.\n");
-		memset(debugfs, 0, 20);
-		strcpy(debugfs, "/tmp/debugfs");
 
-		if (access(debugfs, F_OK)) {
-			debug_dir_existed = mkdir(debugfs, 644);
-			if (debug_dir_existed) {
-				fprintf(stderr,
-					"Can't create %s mount point.\n",
-					debugfs);
-				free(debugfs);
-				exit(-1);
-			}
+		memset(debugfs, 0, PATH_MAX);
+		sprintf(debugfs, "/tmp/debugfs.XXXXXX");
+
+		if (!mkdtemp(debugfs)) {
+			fprintf(stderr,
+				"Can't create %s mount point.\n",
+				debugfs);
+			free(debugfs);
+			exit(-1);
 		}
+
 		rc = mount("none", debugfs, "debugfs", 0, NULL);
 		if (rc) {
 			fprintf(stderr,
@@ -152,7 +138,7 @@ print_lockdump(int argc, char **argv)
 		file = fopen(path, "rt");
 		if (file) {
 			while (fgets(line, PATH_MAX, file)) {
-				printf(line);
+				printf("%s", line);
 			}
 			fclose(file);
 		} else {
@@ -166,10 +152,7 @@ print_lockdump(int argc, char **argv)
 	/* Check if we mounted the debugfs and if so, unmount it. */
 	if (!rc) {
 		umount(debugfs);
-		/* Check if we created the debugfs mount point and if so,
-		   delete it. */
-		if (!debug_dir_existed)
-			rmdir(debugfs);
+		rmdir(debugfs);
 	}
 	free(debugfs);
 }
@@ -198,16 +181,12 @@ print_flags(struct gfs2_dinode *di)
 {
 	if (di->di_flags) {
 		printf("Flags:\n");
-		if (di->di_flags & GFS2_DIF_SYSTEM)
-			printf("  system\n");
 		if (di->di_flags & GFS2_DIF_JDATA)
 			printf("  jdata\n");
 		if (di->di_flags & GFS2_DIF_EXHASH)
 			printf("  exhash\n");
 		if (di->di_flags & GFS2_DIF_EA_INDIRECT)
 			printf("  ea_indirect\n");
-		if (di->di_flags & GFS2_DIF_DIRECTIO)
-			printf("  directio\n");
 		if (di->di_flags & GFS2_DIF_IMMUTABLE)
 			printf("  immutable\n");
 		if (di->di_flags & GFS2_DIF_APPENDONLY)
@@ -228,12 +207,10 @@ print_flags(struct gfs2_dinode *di)
 static unsigned int 
 get_flag_from_name(char *name)
 {
-	if (strncmp(name, "system", 6) == 0)
-		return GFS2_DIF_SYSTEM;
-	else if (strncmp(name, "jdata", 5) == 0)
+	if (strncmp(name, "jdata", 5) == 0)
 		return FS_JOURNAL_DATA_FL;
-	else if (strncmp(name, "directio", 8) == 0)
-		return FS_DIRECTIO_FL;
+	else if (strncmp(name, "exhash", 6) == 0)
+		return FS_INDEX_FL;
 	else if (strncmp(name, "immutable", 9) == 0)
 		return FS_IMMUTABLE_FL;
 	else if (strncmp(name, "appendonly", 10) == 0)
@@ -453,9 +430,7 @@ print_journals(int argc, char **argv)
 	if (sbd.device_fd < 0)
 		die("can't open device %s: %s\n",
 		    sbd.device_name, strerror(errno));
-	if (!find_gfs2_meta(&sbd))
-		mount_gfs2_meta(&sbd);
-	lock_for_admin(&sbd);
+	mount_gfs2_meta(&sbd);
 
 	sprintf(jindex_name, "%s/jindex", sbd.metafs_path);
 	jindex = opendir(jindex_name);
